'
' -----[ Program Description ]----------------------------------------
'
' This program implements the Home protocol in BASCOM for an universal I/O device
' It uses STK200 development kit for testing purpouse.
' It's based on XOR CRC as error detection method to give secure data transfer.
'
' -----[ Disclaimer ]-------------------------------------------------
'
' This example is offered on an "AS IS" basis, no warranty expressed or implied.
' The programers disclaim liability of any damages associated with the use of the hardware or software
' described herein. You use it on your own risk. We are not able to provide any free support.
'
' Copyright (c) 1999 Mike Eitel all rights reserved
'
' -----[ Revision History ]-------------------------------------------
'
' 991220 - Version 1.00AVR Ver 1.10 with  commands for AVR 2313   Mike Eitel
'
' -----[ Aliases ]----------------------------------------------------

'$regfile = "2313def.dat"

' Define serial communication parameters
$baud = 9600
      '$crystal = 8000000                                            ' Frequency of fast STK 200
      'Ubrr = 51                                                     ' For 9600 baud
$crystal = 4000000                                            ' Frequency of STK 200
Ubrr = 25                                                     ' For 9600 baud
'Ucr = &B11011000                                              ' Allow receiver and transmitter ISR
'
' -----[ Constants ]--------------------------------------------------
Const Sync = &B01010101                                       ' Synchronisation byte

Const Telegrambytes = 9                                       ' Number of transmitted bytes

Const Myaddress = &H04                                        ' Address for this node (1-255)

Const Cmaxchar = 12                                           ' Number of characters for serial buffer

Const Dataarea = 12                                           ' Size of the Databuffer

' -----[ Variables ]--------------------------------------------------
'
Dim Crc As Byte                                               ' CRC variable for comunication
Dim Db(telegrambytes) As Byte                                 ' Packet data ( 2-5 ) with command (1)
Dim Sdbcounter As Byte                                        ' Lenght of send telegram
Dim Sdbcountertmp As Byte                                     ' Helper for send telegram
Dim Senderbit As Byte                                         ' Helper byte for bit 6 of portd

Dim Hisaddress As Byte                                        ' Byte containing the last requester adress
'                            Commands
'                         +-----------------+                   W    = 0/1 Read / Write
'                         | W C C C A A A A |                   CCC  = 0. 7  Different command categories
'                         +-----------------+                        = 7 prioritised commands
'                                                                    = 6
'                                                                    = 5
'                                                                    = 4
'                                                                    = 3
'                                                                    = 2 EPROM Data Area
'                                                                    = 1 SRAM Data Area
'                                                                    = 0
'                                                               AAAA = 0..15 Pointer for values
'
'Dim Commands As Byte                                          ' Byte to recognize the requested commands
Dim Datas(dataarea) As Byte                                   ' Place to keep the dynamic datas

Dim Bitsreceived As Byte                                      ' Byte for signalling several received characters
Dim Bitbuffer(cmaxchar) As Byte                               ' Receive serial-buffer
Dim Bitcounter As Byte                                        ' Byte counter of received characters
Dim Bitreader As Byte                                         ' Byte counter of transformed characters

Dim Multiplexa As Byte                                        ' Counter to multiplex 32 In- and 32 Outputpins
Dim Multiplexb As Byte                                        ' Helper for multiplexer
Dim Multiplexc As Byte                                        ' Helper pointer for fast I/O reaction
Dim Multiplexd As Byte                                        ' Helper for fast I/O reaction

' -----[ Initialization ]---------------------------------------------
'
' Define port parameters
'  Port B  is used for data exchange with the latches
'  Ddrb = &HFF                                                 ' Switch to WRITE to multiple multiple I/O Data port
'  Ddrb = &H00                                                 ' Switch to READ from multiple I/O Data port

'  Port D  is used for communication and for latch controll
  Ddrd = &B01111100                                           ' Set Portd Pin 2..6 as Outputs
'  Ddrd.0                                                      ' RXD
'  Ddrd.1                                                      ' TXD
'  Ddrd.2                                                      ' Strobe for output latches Allegro UCN 5801A
'  Ddrd.3                                                      ' Low bit of adress decoder 1..8
'  Ddrd.4                                                      ' Middle bit of adress decoder
'  Ddrd.5                                                      ' High bit of adress decoder high = output chips
'  Ddrd.6                                                      ' Enable for RS422 send pin
'  Ddrd.7                                                      ' NC for 2313

  ' Setup interrupts to defined startup
  Ucr = &B11011000                                            ' Allow receiver and transmitter ISR
  'Enable Utxc                                                ' Disable send isr
  'Enable Urxc                                                ' Enable receive isr
  'Disable Udre                                               ' Enable empty send buffer isr
  On Urxc Rec_isr                                             ' Define serial receive ISR
  On Utxc Send_isr                                            ' Define serial send ISR
  On Udre Ude_isr Nosave                                      ' Define empty serial buffer ISR
  Enable Serial                                               ' Allow protokoll treatment starts

  Config Timer0 = Timer , Prescale = 64                       ' Define update speed
  On Timer0 Timer0_isr                                        ' Define ISR when timer 0 overflows                                         '
  Enable Timer0                                               ' Allow I/O treating

  Multiplexa = 0                                              ' Set to defined Startup value
  Sdbcounter = 0                                              ' Set to defined Startup value

  Enable Interrupts                                           ' Start the system timing from this line on


'------[ Program ]----------------------------------------------------
'
_com_z:
                               ' Startingpoint for telegrams
   If Bitsreceived >= 1 Then                                  ' Received something
      Bitreader = Bitcounter - Bitsreceived
      Incr Bitreader
      Db(1) = Bitbuffer(bitreader)                            ' transfer into receive telegram buffer
   Else
      Goto _com_z                                             ' give task switching a chance
   End If
   Decr Bitsreceived

   ' If received data is SYNC byte read next 8 bytes from master, if not return to start
   If Db(1) <> Sync Then
      Goto _com_z                                             ' give task switching a chance
   End If

   Incr Bitreader
   If Bitreader >= Cmaxchar Then                              ' Buffer full = will generate unread message
      Bitreader = 1                                           ' Rewrap character serial counter
   End If

_com_1:
   If Bitsreceived < 8 Then
      Goto _com_1                                             ' give task switching a chance
   End If

   ' Get telegram in binary mode
   Db(2) = Bitbuffer(bitreader)                               ' DAD
   Incr Bitreader
   Db(3) = Bitbuffer(bitreader)                               ' SAD
   Incr Bitreader
   Db(4) = Bitbuffer(bitreader)                               ' CMD
   Incr Bitreader
   Db(5) = Bitbuffer(bitreader)                               ' DB1
   Incr Bitreader
   Db(6) = Bitbuffer(bitreader)                               ' DB2
   Incr Bitreader
   Db(7) = Bitbuffer(bitreader)                               ' DB3
   Incr Bitreader
   Db(8) = Bitbuffer(bitreader)                               ' DB4
   Incr Bitreader
   Db(9) = Bitbuffer(bitreader)                               ' CRC
   Incr Bitreader

   Bitcounter = 1                                             ' Reset for next telegram
   Bitsreceived = 0                                           ' Reset for next telegram

   ' Address check routine checks if this is the node addressed, if not goto Startingpoint
   If Db(2) <> Myaddress Then
      Goto _com_1:
   End If

   Hisaddress = Db(3)                                         ' keep in mind his address

   ' CRC check routine and calculate CRC for all received bytes
   Crc = Sync Xor Db(2)
   Crc = Crc Xor Db(3)
   Crc = Crc Xor Db(4)
   Crc = Crc Xor Db(5)
   Crc = Crc Xor Db(6)
   Crc = Crc Xor Db(7)
   Crc = Crc Xor Db(8)

   ' Check if there was any CRC errors, if so send NAK
   If Crc <> Db(9) Then
'      Gosub _nak                                              ' Gosub _nak Routine
      Goto _com_z
   End If

_com_4:
   ' No CRC errors in packet so check what to do and then start writing data or do according commands
'   Commands = Db(4)
   Select Case Db(4)                                          ' Decide what kind of command categorie to treat
         Case 16 :                                            ' Reading datas 1..4
                    Db(5) = Datas(1)                          ' Reading
                    Db(6) = Datas(2)                          ' Reading
                    Db(7) = Datas(3)                          ' Reading
                    Db(8) = Datas(4)                          ' Reading
                    Gosub _send                               ' Normal re-reading SRAM
         Case 144 :                                           ' Writing Datas 1..4
                    Datas(1) = Db(5)                          ' Writing
                    Datas(2) = Db(6)                          ' Writing
                    Datas(3) = Db(7)                          ' Writing
                    Datas(4) = Db(8)                          ' Writing
                    Gosub _ack                                ' End command with ACK = normal re-reading SRAM
         Case 145 :                                           ' Writing datas 5..8
                    Datas(5) = Db(5)                          ' Writing
                    Datas(6) = Db(6)                          ' Writing
                    Datas(7) = Db(7)                          ' Writing
                    Datas(8) = Db(8)                          ' Writing
                    Gosub _ack                                ' End command with ACK = normal re-reading SRAM
         Case 146 :                                           ' Writing datas 9..12
                    Datas(9) = Db(5)                          ' Writing
                    Datas(10) = Db(6)                         ' Writing
                    Datas(11) = Db(7)                         ' Writing
                    Datas(12) = Db(8)                         ' Writing
                    Gosub _ack                                ' Normal re-reading SRAM
         Case 254 :                                           ' Spezial / priorised commands
                    Disable Timer0                            ' Disallow cyclic send of Datas 1..4
'                    Gosub _ack
         Case 255 :
                    Enable Timer0                             ' Allow cyclic send of Datas 1..4
'                    Gosub _ack                                ' Command was sucessfull
         Case Else :
'                    Gosub _nak                                ' NOT recognized command
  End Select

  ' Done, go back to Start and wait for a new packet
   Goto _com_z


' End of the main program
' -----[ End ]--------------------------------------------------------



' -----[ Subroutines ]------------------------------------------------
'
' Send NAK (i.e tell master that packet was bad)
'_nak:
'   Db(4) = 111
'   Db(5) = 255
'   Db(6) = 255
'   Db(7) = 255
'   Db(8) = 255
'   Gosub _send
'Return

' Send Ack (i.e Tell Master That Packet Was Ok)
_ack:
   Db(4) = Db(4) And &H7F                                     ' Just convert Send to Receive command
   Gosub _send
Return

' Subroutine for sending values -> Sync,DAD,SAD,COM,Db1,Db2,Db3,Db4,CRC,
_send:
'   Db(1) = Sync                                               ' Sync
   Db(2) = Hisaddress                                         ' Exchange SAD and DAD address bytes
   Db(3) = Myaddress                                          ' Put DAD in tempory variable
   Crc = Sync Xor Db(2)                                       '
   Crc = Crc Xor Db(3)                                        '
   Crc = Crc Xor Db(4)                                        '      XOR calculated CRC byte
   Crc = Crc Xor Db(5)                                        '
   Crc = Crc Xor Db(6)                                        '
   Crc = Crc Xor Db(7)                                        '
   Crc = Crc Xor Db(8)                                        '
   Db(9) = Crc                                                ' Put CRC value to complete telegram

    'Send Packet To Master , Including The Sync Byte
_again:
    If Sdbcounter < 1 Then                                    ' Pure safety loop in case of to much sending
       Set Portd.6                                            ' Set switch to send modus pin
       Senderbit = &H40                                       ' Set the bit for portd.6 in timer0 cycle
       Sdbcounter = 9                                         ' Set buffer to full lenght
       Udr = Sync                                             ' Start ISR work by Sync byte sending
    Else
       Goto _again                                            ' Safety loop should never run
    End If
Return


' -----[ Interrupt routines ]------------------------------------------------

' Interrupt routine for preparing serial port
' On powerup it's treated once !
' This routine should not be needed but it only works with this routine.
Ude_isr:
      Ucr = &B11011000                                        ' Number of character flag
Return


' Interrupt routine for reading serial input
Rec_isr:
   If Bitcounter < Cmaxchar Then                              ' Does it fit into the buffer?
      Incr Bitcounter                                         ' Increase buffer counter
   Else
      Bitcounter = 1                                          ' Safety rewrap character serial counter
   End If
   Bitbuffer(bitcounter) = Udr                                ' Add to buffer array
   Incr Bitsreceived
Return


' Interrupt routine for sendig serial output
Send_isr:
   If Sdbcounter > 0 Then                                     ' (0) = Dirty trick to see buffer size ?
      Sdbcountertmp = 11 - Sdbcounter
      Udr = Db(sdbcountertmp)                                 ' Increase buffer counter
      Decr Sdbcounter
   Else
      Reset Portd.6                                           ' Clean switch to send modus pin
      Senderbit = &H00                                        ' Clear the bit for portd.6 in timer0 cycle
   End If
Return


' Interrupt routine for multiplexing In / Outputs
Timer0_isr:                                                   ' Controlls the multiplexing of data versus I/O's
    Multiplexb = Multiplexa                                   ' Multiplexa on this line 0..15
    Shift Multiplexb , Left , 2                               ' To handle the UART bits shift
    Portd = Multiplexb Or Senderbit                           '    and trick to get a strobe impuls on pin 2

    Multiplexb = Multiplexa                                   ' Multiplexb to 1..16
    Incr Multiplexa                                           ' Cycle counter for the datas 1..16

    Shift Multiplexb , Right , 1                              ' Multiplexa down to 0..7
    Multiplexc = Multiplexb                                   ' Multiplexc to 0..7
    Incr Multiplexb                                           ' Datas ->   1..8

    Shift Multiplexc , Right , 1                              ' Multiplexc to 0..3 base
    Incr Multiplexc                                           ' Dataspointer ->   1..4

    Multiplexd = Multiplexc                                   ' Temp to 1..4
    Set Multiplexd.3                                          ' Add 8 to multiplexd for 9..12

    If Multiplexb < 5 Then                                    ' 1..4 and 4 ..8

       ' ***** Only for testing. Normaly DDRB is &H00 for reading. Also normal instruction is reading !!!  *******
       ' ***** Ddrb = &H00                                           ' Switch to READ from multiple I/O Data port
       ' ***** Datas(multiplexb)= Portb                              ' Read the data from I/O
       ' WRITE lines for testing
       ' ***** Ddrb = &HFF                                            ' Switch to WRITE to multiple multiple I/O Data port
       ' ***** Portb = Datas(multiplexb)                              ' Transfer the data to I/O

       Ddrb = &H00                                            ' Switch to READ from multiple I/O Data port
       Ddrb = &H00                                            ' Time for HW
       Ddrb = &H00                                            ' Time for HW
       Ddrb = &H00                                            ' Time for HW
       Datas(multiplexb) = Portb                              ' Read the data from I/O
    Else
       Ddrb = &HFF                                            ' Switch to WRITE to multiple I/O Data port
       Multiplexc = Not Datas(multiplexc)                     ' If input pin of 1..4 is NOT set AND 9..12 is requesting high
              Multiplexd = Datas(multiplexd) And Multiplexc   '    ,then set according output pin OR when normal command
       Portb = Multiplexd Or Datas(multiplexb)                '    5..8 are set. Then transfer the data to I/O
    End If

    Multiplexa = Multiplexa And &H0F                          ' To cut down cycle counter to 0 ..15
Return